@@ -19,7 +19,6 @@
 
 from io import BytesIO
 import stat
-from unittest import SkipTest
 
 
 from dulwich.object_store import (
@@ -34,6 +33,7 @@ from dulwich.repo import (
     MemoryRepo,
     )
 from dulwich.tests import (
+    SkipTest,
     TestCase,
     )
 from dulwich.tests.utils import (
@@ -56,25 +56,25 @@ class GitFastExporterTests(TestCase):
 
     def test_emit_blob(self):
         b = Blob()
-        b.data = "fooBAR"
+        b.data = b"fooBAR"
         self.fastexporter.emit_blob(b)
-        self.assertEqual('blob\nmark :1\ndata 6\nfooBAR\n',
+        self.assertEqual(b'blob\nmark :1\ndata 6\nfooBAR\n',
             self.stream.getvalue())
 
     def test_emit_commit(self):
         b = Blob()
-        b.data = "FOO"
+        b.data = b"FOO"
         t = Tree()
-        t.add("foo", stat.S_IFREG | 0o644, b.id)
+        t.add(b"foo", stat.S_IFREG | 0o644, b.id)
         c = Commit()
-        c.committer = c.author = "Jelmer <jelmer@host>"
+        c.committer = c.author = b"Jelmer <jelmer@host>"
         c.author_time = c.commit_time = 1271345553
         c.author_timezone = c.commit_timezone = 0
-        c.message = "msg"
+        c.message = b"msg"
         c.tree = t.id
         self.store.add_objects([(b, None), (t, None), (c, None)])
-        self.fastexporter.emit_commit(c, "refs/heads/master")
-        self.assertEqual("""blob
+        self.fastexporter.emit_commit(c, b"refs/heads/master")
+        self.assertEqual(b"""blob
 mark :1
 data 3
 FOO
@@ -84,7 +84,7 @@ author Jelmer <jelmer@host> 1271345553 +0000
 committer Jelmer <jelmer@host> 1271345553 +0000
 data 3
 msg
-M 644 1 foo
+M 644 :1 foo
 """, self.stream.getvalue())
 
 
@@ -103,30 +103,30 @@ class GitImportProcessorTests(TestCase):
     def test_reset_handler(self):
         from fastimport import commands
         [c1] = build_commit_graph(self.repo.object_store, [[1]])
-        cmd = commands.ResetCommand("refs/heads/foo", c1.id)
+        cmd = commands.ResetCommand(b"refs/heads/foo", c1.id)
         self.processor.reset_handler(cmd)
-        self.assertEqual(c1.id, self.repo.get_refs()["refs/heads/foo"])
+        self.assertEqual(c1.id, self.repo.get_refs()[b"refs/heads/foo"])
 
     def test_commit_handler(self):
         from fastimport import commands
-        cmd = commands.CommitCommand("refs/heads/foo", "mrkr",
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            "FOO", None, [], [])
+        cmd = commands.CommitCommand(b"refs/heads/foo",  b"mrkr",
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            b"FOO", None, [], [])
         self.processor.commit_handler(cmd)
         commit = self.repo[self.processor.last_commit]
-        self.assertEqual("Jelmer <jelmer@samba.org>", commit.author)
-        self.assertEqual("Jelmer <jelmer@samba.org>", commit.committer)
-        self.assertEqual("FOO", commit.message)
+        self.assertEqual(b"Jelmer <jelmer@samba.org>", commit.author)
+        self.assertEqual(b"Jelmer <jelmer@samba.org>", commit.committer)
+        self.assertEqual(b"FOO", commit.message)
         self.assertEqual([], commit.parents)
         self.assertEqual(432432432.0, commit.commit_time)
         self.assertEqual(432432432.0, commit.author_time)
         self.assertEqual(3600, commit.commit_timezone)
         self.assertEqual(3600, commit.author_timezone)
-        self.assertEqual(commit, self.repo["refs/heads/foo"])
+        self.assertEqual(commit, self.repo[b"refs/heads/foo"])
 
     def test_import_stream(self):
-        markers = self.processor.import_stream(BytesIO("""blob
+        markers = self.processor.import_stream(BytesIO(b"""blob
 mark :1
 data 11
 text for a
@@ -140,31 +140,31 @@ M 100644 :1 a
 
 """))
         self.assertEqual(2, len(markers))
-        self.assertTrue(isinstance(self.repo[markers["1"]], Blob))
-        self.assertTrue(isinstance(self.repo[markers["2"]], Commit))
+        self.assertTrue(isinstance(self.repo[markers[b"1"]], Blob))
+        self.assertTrue(isinstance(self.repo[markers[b"2"]], Commit))
 
     def test_file_add(self):
         from fastimport import commands
-        cmd = commands.BlobCommand("23", "data")
+        cmd = commands.BlobCommand(b"23", b"data")
         self.processor.blob_handler(cmd)
-        cmd = commands.CommitCommand("refs/heads/foo", "mrkr",
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            "FOO", None, [], [commands.FileModifyCommand("path", 0o100644, ":23", None)])
+        cmd = commands.CommitCommand(b"refs/heads/foo", b"mrkr",
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            b"FOO", None, [], [commands.FileModifyCommand(b"path", 0o100644, b":23", None)])
         self.processor.commit_handler(cmd)
         commit = self.repo[self.processor.last_commit]
         self.assertEqual([
-            ('path', 0o100644, '6320cd248dd8aeaab759d5871f8781b5c0505172')],
+            (b'path', 0o100644, b'6320cd248dd8aeaab759d5871f8781b5c0505172')],
             self.repo[commit.tree].items())
 
     def simple_commit(self):
         from fastimport import commands
-        cmd = commands.BlobCommand("23", "data")
+        cmd = commands.BlobCommand(b"23", b"data")
         self.processor.blob_handler(cmd)
-        cmd = commands.CommitCommand("refs/heads/foo", "mrkr",
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            "FOO", None, [], [commands.FileModifyCommand("path", 0o100644, ":23", None)])
+        cmd = commands.CommitCommand(b"refs/heads/foo", b"mrkr",
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            b"FOO", None, [], [commands.FileModifyCommand(b"path", 0o100644, b":23", None)])
         self.processor.commit_handler(cmd)
         commit = self.repo[self.processor.last_commit]
         return commit
@@ -176,34 +176,34 @@ M 100644 :1 a
         :return: The created commit object
         """
         from fastimport import commands
-        cmd = commands.CommitCommand("refs/heads/foo", "mrkr",
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            ("Jelmer", "jelmer@samba.org", 432432432.0, 3600),
-            "FOO", None, [], file_cmds)
+        cmd = commands.CommitCommand(b"refs/heads/foo", b"mrkr",
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            (b"Jelmer", b"jelmer@samba.org", 432432432.0, 3600),
+            b"FOO", None, [], file_cmds)
         self.processor.commit_handler(cmd)
         return self.repo[self.processor.last_commit]
 
     def test_file_copy(self):
         from fastimport import commands
         self.simple_commit()
-        commit = self.make_file_commit([commands.FileCopyCommand("path", "new_path")])
+        commit = self.make_file_commit([commands.FileCopyCommand(b"path", b"new_path")])
         self.assertEqual([
-            ('new_path', 0o100644, '6320cd248dd8aeaab759d5871f8781b5c0505172'),
-            ('path', 0o100644, '6320cd248dd8aeaab759d5871f8781b5c0505172'),
+            (b'new_path', 0o100644, b'6320cd248dd8aeaab759d5871f8781b5c0505172'),
+            (b'path', 0o100644, b'6320cd248dd8aeaab759d5871f8781b5c0505172'),
             ], self.repo[commit.tree].items())
 
     def test_file_move(self):
         from fastimport import commands
         self.simple_commit()
-        commit = self.make_file_commit([commands.FileRenameCommand("path", "new_path")])
+        commit = self.make_file_commit([commands.FileRenameCommand(b"path", b"new_path")])
         self.assertEqual([
-            ('new_path', 0o100644, '6320cd248dd8aeaab759d5871f8781b5c0505172'),
+            (b'new_path', 0o100644, b'6320cd248dd8aeaab759d5871f8781b5c0505172'),
             ], self.repo[commit.tree].items())
 
     def test_file_delete(self):
         from fastimport import commands
         self.simple_commit()
-        commit = self.make_file_commit([commands.FileDeleteCommand("path")])
+        commit = self.make_file_commit([commands.FileDeleteCommand(b"path")])
         self.assertEqual([], self.repo[commit.tree].items())
 
     def test_file_deleteall(self):
